#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;

namespace HeuristicLab.Clients.Hive.SlaveCore {

  /// <summary>
  /// WcfService class is implemented as a Singleton and works as a communication Layer with the Server
  /// </summary>
  public class WcfService : MarshalByRefObject {
    private static WcfService instance;
    /// <summary>
    /// Getter for the Instance of the WcfService
    /// </summary>
    /// <returns>the Instance of the WcfService class</returns>
    public static WcfService Instance {
      get {
        if (instance == null) {
          instance = new WcfService();
          ServiceLocator.Instance.Username = "hiveslave";
          ServiceLocator.Instance.Password = "hiveslave";
        }
        return instance;
      }
    }

    public DateTime ConnectedSince { get; private set; }
    public NetworkEnum.WcfConnState ConnState { get; private set; }

    private WcfService() {
      ConnState = NetworkEnum.WcfConnState.Disconnected;
    }

    #region Job Methods
    public Job GetJob(Guid jobId) {
      return CallHiveService(s => s.GetJob(jobId));
    }

    public void UpdateJob(Job job) {
      CallHiveService(s => s.UpdateJob(job));
    }

    public Guid AddChildJob(Guid parentJobId, Job job, JobData jobData) {
      return CallHiveService(s => s.AddChildJob(parentJobId, job, jobData));
    }

    public IEnumerable<JobData> GetChildJobs(Guid? parentJobId) {
      return CallHiveService(service => {
        IEnumerable<LightweightJob> msg = service.GetLightweightChildJobs(parentJobId, false, false);

        List<JobData> jobs = new List<JobData>();
        foreach (LightweightJob ljob in msg)
          jobs.Add(service.GetJobData(ljob.Id));

        return jobs;
      });
    }

    public void DeleteChildJobs(Guid jobId) {
      CallHiveService(s => s.DeleteChildJobs(jobId));
    }
    #endregion

    #region JobData Methods
    public JobData GetJobData(Guid jobId) {
      return CallHiveService(s => s.GetJobData(jobId));
    }

    /// <summary>
    /// Uploads the JobData and sets a new jobState (while correctly setting Transferring state)
    /// </summary>
    public void UpdateJobData(Job job, JobData jobData, Guid slaveId, JobState state) {
      CallHiveService(service => {
        service.UpdateJob(job);
        job = service.UpdateJobState(job.Id, JobState.Transferring, slaveId, null, null);
        service.UpdateJobData(job, jobData);
        service.UpdateJobState(job.Id, state, slaveId, null, null);
      });
    }

    public Job UpdateJobState(Guid jobId, JobState jobState, string exception) {
      return CallHiveService(s => s.UpdateJobState(jobId, jobState, ConfigManager.Instance.GetClientInfo().Id, null, exception));
    }
    #endregion

    #region Heartbeat Methods
    public List<MessageContainer> SendHeartbeat(Heartbeat heartbeat) {
      return CallHiveService(s => s.Heartbeat(heartbeat));
    }
    #endregion

    #region Plugin Methods
    public IEnumerable<Plugin> GetPlugins() {
      return CallHiveService(s => s.GetPlugins());
    }

    public IEnumerable<PluginData> GetPluginDatas(List<Guid> pluginIds) {
      return CallHiveService(s => s.GetPluginDatas(pluginIds));
    }
    #endregion

    #region Events
    public event EventHandler Connected;
    private void OnConnected() {
      var handler = Connected;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler<EventArgs<Exception>> ExceptionOccured;
    private void OnExceptionOccured(Exception e) {
      var handler = ExceptionOccured;
      if (handler != null) handler(this, new EventArgs<Exception>(e));
    }
    #endregion

    #region Helpers
    /// <summary>
    /// Connects with the Server, registers the events and fires the Connected (and quiet possibly the ConnectionRestored) Event.
    /// </summary>
    public void Connect(Slave slaveInfo) {
      CallHiveService(service => {
        ConnState = NetworkEnum.WcfConnState.Connected;
        ConnectedSince = DateTime.Now;
        service.Hello(slaveInfo);
        OnConnected();
      });
    }

    /// <summary>
    /// Disconnects the Slave from the Server
    /// </summary>
    public void Disconnect() {
      CallHiveService(service => {
        service.GoodBye(ConfigManager.Instance.GetClientInfo().Id);
        ConnState = NetworkEnum.WcfConnState.Disconnected;
      });
    }

    /// <summary>
    /// Network communication Error Handler - Every network error gets logged and the connection switches to faulted state
    /// </summary>
    /// <param name="e">The Exception</param>
    private void HandleNetworkError(Exception e) {
      ConnState = NetworkEnum.WcfConnState.Failed;
      OnExceptionOccured(e);
    }

    public void CallHiveService(Action<IHiveService> call) {
      try {
        ServiceLocator.Instance.CallHiveService(call);
      }
      catch (Exception ex) {
        HandleNetworkError(ex);
      }
    }

    private T CallHiveService<T>(Func<IHiveService, T> call) {
      try {
        return ServiceLocator.Instance.CallHiveService(call);
      }
      catch (Exception ex) {
        HandleNetworkError(ex);
        return default(T);
      }
    }
    #endregion
  }
}